#!/usr/bin/env python

# author: Youri Lammers
# contact: youri.lammers@uit.no / youri.lammers@gmail.com

# Parse the OBITools results to extract the reads, replication
# markers and taxa present in the multiplex results

# note, this script requires numpy

# import the modules used by the script
import argparse, itertools, re, sys, os
from collections import defaultdict
import numpy


# Retrieve the commandline arguments
parser = argparse.ArgumentParser(description = 'Extract the amplicons.')

parser.add_argument('-t','--tables', metavar='OBITab tables', dest='tables',
			action='append', nargs='+',
			help='List of the input OBITab tables.')
parser.add_argument('-f','--fasta', metavar='Fasta files', dest='ff',
			action='append', nargs='+',
			help='List of fasta files')
parser.add_argument('-mr', metavar='minimum reads', dest='min', default=3, type=int,
			help='The minimum number of reads per barcode and detection (default=3).')
parser.add_argument('-min_rep', metavar='minimum repeats', dest='minrep', default=2, type=int,
			help='The minimum number of repeats per barcode (default=2).')
parser.add_argument('-ts', metavar='tag swap rate', dest='tsr', default=0.01, type=float,
			help='The tag swap rate used for filtering (default=0.01).')
args = parser.parse_args()


def parse_table(path, iden_dict, ratio_dict, index_dict, marker_dict, rep_dict, fullread_dict, amplidict,iden_sample_dict):

	# parse through the obitools tables and extract the marker results

	# get the sample name
	sample = os.path.splitext(os.path.basename(path))[0].split('.')[0]
	
	# create the empty variables for storing the column locations
	sequence_pos = 0
	idenperc=0
	idenname=0
	locations = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
	totcount=0
	speclist=0

	iden_dict[sample]['totalreads'] = 0
	iden_dict[sample]['idenreads'] = 0
	iden_dict[sample]['ontargetreads'] = 0

	repeatcount = defaultdict(int)
	read_dict = defaultdict(lambda: defaultdict(int))
	ratio_dict = defaultdict(lambda: defaultdict(list))

	# parse through the file and get the ratio
	for line in open(path):

		# split the lines into the columns
		line = line.strip().split('\t')

		# if the line is the header, extract the column locations
		if line[0] == 'id':

			# loop through the columns
			for i in range(0,len(line)):

				# store the location of the relevant columns
				if 'sample' in line[i]:

					marker = line[i].replace('sample:','').split('_rpt')[0]
					if len(marker) == 4: marker = marker.replace('Vu_','Vu_0')
					rep = line[i].split('_rpt')[1]
					locations[marker]['data'][rep] = i

			# continue with the next line after storing the header info
			continue

		# loop through the markers and reps
		for marker in locations.keys():
			for rep in locations[marker]['data']:
				readcount = int(line[locations[marker]['data'][rep]])
				read_dict[marker][rep] += readcount

	# calculate the ratios
	for marker in read_dict:
		tot = sum([read_dict[marker][rep] for rep in read_dict[marker]])
		for rep in read_dict[marker]:
			ratio_dict[marker][rep] = float(read_dict[marker][rep])/tot

	# go through the file
	for line in open(path):

		# split the lines into the columns
		line = line.strip().split('\t')

		# if the line is the header, extract the column locations
		if line[0] == 'id':

			# loop through the columns
			for i in range(0,len(line)):

				# store the location of the relevant columns
				if line[i] == 'sequence': sequence_pos = i
				if line[i] == 'best_identity:MultiplexRef': idenperc=i
				if line[i] == 'scientific_name': idenname=i
				if line[i] == 'species_list:MultiplexRef': speclist=i
				if 'sample' in line[i]:

					marker = line[i].replace('sample:','').split('_rpt')[0]
					if len(marker) == 4: marker = marker.replace('Vu_','Vu_0')
					rep = line[i].split('_rpt')[1]
					locations[marker]['data'][rep] = i

				elif 'obiclean_status:' in line[i]:

					marker = line[i].replace('obiclean_status:','').split('_rpt')[0]
					if len(marker) == 4: marker = marker.replace('Vu_','Vu_0')
					rep = line[i].split('_rpt')[1]
					locations[marker]['obiclean'][rep] = i
					
			# continue with the next line after storing the header info
			continue

		# store the individual rep counts
		rep_counts = []

		# loop through the markers
		for marker in locations.keys():

			totrepcount = 0
			idenrepcount = 0
			ontargetrepcount = 0

			# calc the tot count for the tag swap ratio
			totcount = sum([int(line[locations[marker]['data'][rep]]) for rep in locations[marker]['data']])

			# get the repeat count and skip if it is below the minimum value
			repcount = len([line[locations[marker]['data'][rep]] for rep in locations[marker]['data'] if int(line[locations[marker]['data'][rep]]) >= args.min and line[locations[marker]['obiclean'][rep]] != 'i'])

			for rep in locations[marker]['data']:

				# get the read count and skip if below the minimum value
				readcount = int(line[locations[marker]['data'][rep]])
				if readcount < args.min: continue

				# check if the sequence is a head or singleton sequence
				obiclean = line[locations[marker]['obiclean'][rep]]
				if obiclean == 'i': continue

				# calculate the expected read count based on the tag swap ratio
				# and subtract the rep if it is a potential swap
				expected = calc_expected(totcount, ratio_dict[marker][rep], args.tsr)
				if readcount < expected:
					repcount -= 1

			# skip if the repcount is too low
			if repcount < args.minrep: continue

			# loop through the replicate columns for the marker
			for rep in locations[marker]['data']:

				# get the read count and skip if below the minimum value
				readcount = int(line[locations[marker]['data'][rep]])
				if readcount < args.min: continue

				# check if the sequence is a head or singleton sequence
				obiclean = line[locations[marker]['obiclean'][rep]]
				if obiclean == 'i': continue

				# calculate the expected read count
				expected = calc_expected(totcount, ratio_dict[marker][rep], args.tsr)

				# reset if it can be explained by a swap, continue
				if readcount < expected: continue

				# store prefect identifications
				if float(line[idenperc]) == 1:

					# check the speclist, if it matches V. gaultherioids, V. uliginosum,
					# change the iden to V. uliginosum
					if line[speclist] == '[\'Vaccinium gaultherioides\', \'Vaccinium uliginosum\']':
						line[idenname] = 'Vaccinium uliginosum'

					# store the identification
					if line[idenname] not in iden_dict[sample][marker]:
						iden_dict[sample][marker].append(line[idenname])
					if line[idenname] not in iden_dict[sample]['totalspecies']:
						iden_dict[sample]['totalspecies'].append(line[idenname])
					index_dict[line[idenname]] += 1

				# add it to the index_dict
				iden_dict[sample]['totalreads'] += readcount
				iden_sample_dict[sample][marker]['totalreads'] += readcount
				if float(line[idenperc]) == 1:
					iden_dict[sample]['idenreads'] += readcount
					iden_sample_dict[sample][marker]['idenreads'] += readcount
				if float(line[idenperc]) == 1 and line[idenname] == 'Vaccinium uliginosum':
					iden_dict[sample]['ontargetreads'] += readcount
					iden_sample_dict[sample][marker]['ontargetreads'] += readcount
					fullread_dict[sample][marker][line[sequence_pos]] += readcount


				# add the species to the marker
				if line[idenname] not in iden_dict[sample][marker]:
					iden_dict[sample][marker].append(line[idenname])
				
				# keep track of the reads for the marker
				marker_dict[sample][marker] += readcount
				marker_dict['totalreads'][marker] += readcount
				if float(line[idenperc]) == 1:
					marker_dict['idenreads'][marker] += readcount
				if float(line[idenperc]) == 1 and line[idenname] == 'Vaccinium uliginosum':
					marker_dict['ontargetreads'][marker] += readcount
					repeatcount[rep] += readcount

				# store the detected sequences
				if line[sequence_pos] not in amplidict['total'][marker]:
					amplidict['total'][marker].append(line[sequence_pos])
				if line[sequence_pos] not in amplidict['total']['total']:
					amplidict['total']['total'].append(line[sequence_pos])
				
				if line[sequence_pos] not in amplidict['iden'][marker] and float(line[idenperc]) == 1:
					amplidict['iden'][marker].append(line[sequence_pos])
				if line[sequence_pos] not in amplidict['iden']['total'] and float(line[idenperc]) == 1:
					amplidict['iden']['total'].append(line[sequence_pos])
				
				if line[sequence_pos] not in amplidict['target'][marker] and float(line[idenperc]) == 1 and line[idenname] == 'Vaccinium uliginosum':
					amplidict['target'][marker].append(line[sequence_pos])
				if line[sequence_pos] not in amplidict['target']['total'] and float(line[idenperc]) == 1 and line[idenname] == 'Vaccinium uliginosum':
					amplidict['target']['total'].append(line[sequence_pos])


				# update the repcount
				totrepcount += 1
				if float(line[idenperc]) == 1: idenrepcount += 1
				if float(line[idenperc]) == 1 and line[idenname] == 'Vaccinium uliginosum':
					ontargetrepcount += 1

			if totrepcount >= args.minrep:
				rep_dict['totrep'][sample][marker].append(totrepcount)
				rep_dict['totrep'][sample]['totrep'].append(totrepcount)
			if idenrepcount >= args.minrep:
				rep_dict['idenrep'][sample][marker].append(idenrepcount)
				rep_dict['idenrep'][sample]['idenrep'].append(idenrepcount)
			if ontargetrepcount >= args.minrep:
				rep_dict['ontargetrep'][sample][marker].append(ontargetrepcount)
				rep_dict['ontargetrep'][sample]['ontargetrep'].append(ontargetrepcount)

	print '{0}\t{1}'.format(sample,'\t'.join([str(repeatcount[rep]) for rep in ['1','2','3','4']]))

	#return [iden_dict,ratio_list,index_dict]
	return [iden_dict,ratio_dict,index_dict,marker_dict,rep_dict,fullread_dict,amplidict,iden_sample_dict]


def avg(list):
	# produce a rounded average
	if len(list) > 0:
		return round(float(sum(list))/len(list),2)
	else:
		return 0

def calc_expected(tot_count, prop, rate):

	# calculate and return the swap rat
	return tot_count * prop * rate


def extract_sequences(file):

	# open the sequence file submitted by the user
	sequence_file = open(file)

	# parse throught the file
	lines = (x[1] for x in itertools.groupby(sequence_file, key=lambda line: line[0] == '>'))

	# walk through the header and obtain the sequences
	for headers in lines:

		header = headers.next().strip().lower().replace('>>','>').split(';')[0]
		if header == '': continue

		sequence = ''.join(line.strip() for line in lines.next()).lower()

		# yield the header + sequence
		yield [header, sequence]


def get_ref():

	# set up the results dictionary
	ref_dict = defaultdict(lambda: defaultdict(list))
		
	# loop through the files
	for file in args.ff[0]:

		# get a read from the input file
		for seq in extract_sequences(file):

			# only use the sequences of interest
			if re.search(r'>vu_[0-9]{2}_vaccinium_uliginosum', seq[0]):

				# get the sample from the fasta hearder by removing
				# the marker and contig info
				sample = re.sub(r'>vu_[0-9]{2}_','',seq[0].split('.')[0])

				# get the marker
				marker = file.split('/')[1].replace('.fasta','')

				# add the sequence to the results
				ref_dict[marker][seq[1]].append(sample)

	# return the ref sequences
	return ref_dict


def main ():

	groups = {'vaccinium_uliginosum_canada_trom_v_355000_cdm_apu':'Arctic','vaccinium_uliginosum_russia_trom_v_355008_cdm_aza':'Arctic',
				'vaccinium_uliginosum_france_bgn_afb':'A-A','vaccinium_uliginosum_france_bgn_eap':'Alpine',
				'vaccinium_uliginosum_france_trom_v_355006_cdm_auo':'Alpine','vaccinium_uliginosum_france_trom_v_355013_cdm_azf':'Alpine',
				'vaccinium_uliginosum_russia_trom_v_355010_cdm_azc':'Beringian','vaccinium_uliginosum_microphyllum_alta_trom_v_162265_bxa_bns':'A-A',
				'vaccinium_uliginosum_nordkapp_trom_v_355015_cdm_azg':'Arctic','vaccinium_uliginosum_northkorea_trom_v_355004_cdm_apw':'Arctic',
				'vaccinium_uliginosum_canada_trom_v_355009_cdm_azb':'Arctic','vaccinium_uliginosum_senja_trom_v_963388_bxa_bnr':'A-A',
				'vaccinium_uliginosum_spain_trom_v_355001_cdm_apa':'Alpine','vaccinium_uliginosum_spain_trom_v_355005_cdm_apx':'Alpine',
				'vaccinium_uliginosum_greenland_trom_v_355011_cdm_azd':'Arctic','vaccinium_uliginosum_switzerland_trom_v_355003_cdm_apv':'A-A',
				'vaccinium_uliginosum_uliginosum_loppa_trom_v_962025_bmb_fao':'A-A','vaccinium_uliginosum_svalbard_trom_v_355012_cdm_aze':'Arctic'}

	ref_dict = get_ref()



	# set up the results dictionary and the min max lenght variables
	iden_dict = defaultdict(lambda: defaultdict(list))
	iden_sample_dict = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
	ratio_dict = defaultdict(list)
	index_dict = defaultdict(int)
	marker_dict = defaultdict(lambda: defaultdict(int))
	rep_dict = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
	fullread_dict = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
	amplidict = defaultdict(lambda: defaultdict(list))

	# print the number of on target reads for each repeat
	print 'sample\ton-target rep1\ton-target rep2\ton-target rep3\ton-target rep4'		
	for table in args.tables[0]:
		results = parse_table(table, iden_dict, ratio_dict, index_dict, marker_dict,rep_dict,fullread_dict,amplidict,iden_sample_dict)
		iden_dict, ratio_dict, index_dict, marker_dict, rep_dict,fullread_dict, amplidict, iden_sample_dict = results
	print '\n\n\n'

	print 'sample\ttotal reads\ttotal taxa\ttaxa list'
	for sample in sorted(iden_dict.keys()):
		print '{0}\t{1}\t{2}\t{3}'.format(sample,iden_dict[sample]['totalreads'],len(iden_dict[sample]['totalspecies']),', '.join(iden_dict[sample]['totalspecies']))
	print '\n\n\n'

	print 'taxa\tdetections in dataset (all detections for all markers and samples)'
	for i in index_dict:
		print '{0}\t{1}'.format(i, index_dict[i])
	print '\n\n\n'

	totalspeclist = []

	# calc total markers present and on target markers present
	print 'sample\ttotal markers\ton-target markers'
	for sample in sorted(iden_dict.keys()):
		for i in iden_dict[sample]:
			if 'Vu_' in i:
				for j in iden_dict[sample][i]:
					totalspeclist.append(j)
		print '{0}\t{1}\t{2}'.format(sample,
			len([key for key in iden_dict[sample].keys() if 'Vu_' in key]),
			len([key for key in iden_dict[sample].keys() if 'Vu_' in key and 'Vaccinium uliginosum' in iden_dict[sample][key]]))
	print '\n\n\n'

	# print the total taxa
	print 'number of taxa detected {0}\n\n\n'.format(len(list(set(totalspeclist))))
	
	# get the number of markers, both on and off target

	# get the markers present
	markers = [[mark for mark in iden_dict[sample] if 'Vu' in mark] for sample in iden_dict]
	markers = [m for i in markers for m in i]
	markers = sorted(list(set(markers)))


	# loop through the reference dict to get the number of haplogroups per marker
	print 'marker\tnumber of unique haplogroups'
	for marker in markers:
		groupdict = defaultdict(list)
		groupcount = 0
		grouplist = ['Arctic','Alpine','A-A','Beringian']
		for seq in ref_dict[marker]:
			for ref in ref_dict[marker][seq]:
				group = groups[ref]
				groupdict[seq].append(group)
		for group in grouplist:
			unique = 1
			for seq in groupdict:
				groupseq = list(set(groupdict[seq]))
				if group in groupseq and len(groupseq) > 1:
					unique = 0
			if unique == 1:
				groupcount += 1
		print '{0}\t{1}'.format(marker,groupcount)
	print '\n\n\n'


	samples = [sample for sample in iden_dict if 'Vaccinium uliginosum' in iden_dict[sample]['totalspecies']]


	# score the marker occurences
	print 'marker\tnum in sedadna\ton-target num in sedadna'
	for mark in markers:
		print '{0}\t{1}\t{2}'.format(mark,
			len([iden_dict[i][mark] for i in samples if len(iden_dict[i][mark]) >= 1]),
			len([iden_dict[i][mark] for i in samples if len(iden_dict[i][mark]) >= 1 and 'Vaccinium uliginosum' in iden_dict[i][mark]]))
	print '\n\n\n'

	# print the marker read counts
	print 'marker\ttotal reads\tiden reads\ton-target reads'
	for mark in markers:
		print '{0}\t{1}\t{2}\t{3}'.format(mark,
			marker_dict['totalreads'][mark],marker_dict['idenreads'][mark],
			marker_dict['ontargetreads'][mark])
	print '\n\n\n'

	# print readcounts
	print 'sample\ttotal count\tafter iden\ton-target'
	for sample in sorted(iden_dict.keys()):
		print '{0}\t{1}\t{2}\t{3}'.format(sample,iden_dict[sample]['totalreads'],
			iden_dict[sample]['idenreads'],iden_dict[sample]['ontargetreads'])
	print '\n\n\n'

	# print average repeat counts
	print 'sample\ttotal rep avg\tiden rep avg\ton-target rep avg'
	for sample in sorted(iden_dict.keys()):
		print '{0}\t{1}\t{2}\t{3}'.format(sample,avg(rep_dict['totrep'][sample]['totrep']),
			avg(rep_dict['idenrep'][sample]['idenrep']),
			avg(rep_dict['ontargetrep'][sample]['ontargetrep']))
	print '\n\n\n'

	# print rep counts and number of sequences
	print 'marker\ttotal rep count\tiden rep\ton-target rep\ttotal seq\tiden seq\ton-target seq'
	for mark in markers:
		print '{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}'.format(mark,
			avg([i for sample in iden_dict.keys() for i in rep_dict['totrep'][sample][mark]]),
			avg([i for sample in iden_dict.keys() for i in rep_dict['idenrep'][sample][mark]]),
			avg([i for sample in iden_dict.keys() for i in rep_dict['ontargetrep'][sample][mark]]),
			len(amplidict['total'][mark]),
			len(amplidict['iden'][mark]),
			len(amplidict['target'][mark]))

	totvarlist = [rep_dict['totrep'][x]['totrep'][z] for x in rep_dict['totrep'] for z in range(0, len(rep_dict['totrep'][x]['totrep']))]
	idenvarlist = [rep_dict['idenrep'][x]['idenrep'][z] for x in rep_dict['idenrep'] for z in range(0, len(rep_dict['idenrep'][x]['idenrep']))]
	ontarvarlist = [rep_dict['ontargetrep'][x]['ontargetrep'][z] for x in samples for z in range(0, len(rep_dict['ontargetrep'][x]['ontargetrep']))]
	print 'type\tavg\tsd\ttotal mark'
	print '{0}\t{1}\t{2}\t{3}'.format('totrep',avg(totvarlist),numpy.std(totvarlist),len(amplidict['total']['total']))
	print '{0}\t{1}\t{2}\t{3}'.format('idenrep',avg(idenvarlist),numpy.std(idenvarlist),len(amplidict['iden']['total']))
	print '{0}\t{1}\t{2}\t{3}'.format('ontargetrep',avg(ontarvarlist),numpy.std(ontarvarlist),len(amplidict['target']['total']))
	print '\n\n\n'
	
	# generate the sample read counts
	print 'sample\ttot read\tiden read\ton-target\tper on-target'
	for sample in sorted(iden_dict.keys()):
		totread = sum([iden_sample_dict[sample][mark]['totalreads'] for mark in markers])
		idenread = sum([iden_sample_dict[sample][mark]['idenreads'] for mark in markers])
		ontargetread = sum([iden_sample_dict[sample][mark]['ontargetreads'] for mark in markers])
		try:
			perontarget = (float(ontargetread)/idenread)*100
		except:
			perontarget = 0
		print '{0}\t{1}\t{2}\t{3}\t{4}'.format(sample,totread,idenread,ontargetread,perontarget)
	print '\n\n\n'

	# get the average on target reads for each sample
	print 'sample\tavg on-target reads'
	for sample in sorted(iden_dict.keys()):
		perlist = []
		for mark in markers:
			if iden_sample_dict[sample][mark]['ontargetreads'] > 0:
				perlist.append((float(iden_sample_dict[sample][mark]['ontargetreads'])/iden_sample_dict[sample][mark]['idenreads'])*100)
		print '{0}\t{1}'.format(sample,avg(perlist))
	print '\n\n\n'

	# generate the marker read counts
	print 'marker\tcount sedadna\tcount\ttot read\tiden read\ton-target\tper on-target'
	for mark in markers:
		totread = sum([iden_sample_dict[sample][mark]['totalreads'] for sample in sorted(iden_dict.keys())])
		idenread = sum([iden_sample_dict[sample][mark]['idenreads'] for sample in sorted(iden_dict.keys())])
		ontargetread = sum([iden_sample_dict[sample][mark]['ontargetreads'] for sample in sorted(iden_dict.keys())])
		count = len([iden_sample_dict[sample][mark]['ontargetreads'] for sample in sorted(iden_dict.keys()) if iden_sample_dict[sample][mark]['ontargetreads'] > 0])
		idencount = len([iden_sample_dict[sample][mark]['ontargetreads'] for sample in sorted(iden_dict.keys()) if iden_sample_dict[sample][mark]['idenreads'] > 0])
		try:
			perontarget = round((float(ontargetread)/idenread)*100,2)
		except:
			perontarget = 0
		print '{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}'.format(mark,idencount,count,totread,idenread,ontargetread,perontarget)
	print '\n\n\n'

	# get the average on target reads for each marker per lake
	print 'marker\tavg on-target per lake'
	for mark in markers:
		perlist = []
		for sample in sorted(iden_dict.keys()):
			if iden_sample_dict[sample][mark]['ontargetreads'] > 0:
				perlist.append((float(iden_sample_dict[sample][mark]['ontargetreads'])/iden_sample_dict[sample][mark]['idenreads'])*100)
		print '{0}\t{1}'.format(mark,avg(perlist))


if __name__ == '__main__':
	main()